#ifndef LSSDK_ENFORCE_SIMPLE_HPP_INCLUDED
#define LSSDK_ENFORCE_SIMPLE_HPP_INCLUDED

#if _MSC_VER >= 1020
#  pragma once
#endif // _MSC_VER

//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
//  enforce_simple.hpp
//  This is a part of the LS C/C++ SDK source code.
//
//  Copyright (C) 2003-2004 ilmcuts.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-

#include <stdexcept>


namespace lssdk
{

//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// default policies
//
struct DefaultPredicate
{
    template <class T> static bool Wrong(const T& obj)
    {
        return !obj;
    }
};

template <class Exception = std::runtime_error>
struct DefaultRaiser
{
    template <class T> static void Throw(const T&, const char* pszWhat)
    {
        throw Exception(pszWhat);
    }
};


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Enforce
//
template <class Predicate, class Raiser, class T>
T& Enforce(T& t, const char* pszWhat)
{
    if (Predicate::Wrong(t))
    {
        Raiser::Throw(t, pszWhat);
    }
    
    return t;
}

template <class Predicate, class Raiser, class T>
const T& Enforce(const T& t, const char* pszWhat)
{
    if (Predicate::Wrong(t))
    {
        Raiser::Throw(t, pszWhat);
    }
    
    return t;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ENFORCE macros
//
#define ENFORCE lssdk::Enforce<lssdk::DefaultPredicate, lssdk::DefaultRaiser<> >
#define ENFORCE_ARG lssdk::Enforce<lssdk::DefaultPredicate, lssdk::DefaultRaiser<std::invalid_argument> >


} // namespace toolkit

#endif // !defined(LSSDK_ENFORCE_SIMPLE_HPP_INCLUDED)
